#!/usr/bin/env ruby

base_dir = File.dirname(__FILE__)

require "./#{base_dir}/sparql.rb"
require 'json'
require 'erb'
require 'fileutils'
require 'tempfile'
require 'open-uri'

ISQL = '/data/store/virtuoso7.2hv/bin/isql 20831 dba dba'
ISQL_OPT = 'VERBOSE=OFF BANNER=OFF PROMPT=OFF ECHO=OFF BLOBS=ON ERRORS=stderr'
VARIANT_DIR = '/data/store/rdf/variant'
OUTPUT_DIR = "#{VARIANT_DIR}/clinvar/current/additional_data"
class ClinVarAdditionalData

  def initialize
    FileUtils.mkdir_p("#{OUTPUT_DIR}")
    @base_dir = File.dirname(__FILE__)
    @complex_variant_list_file = "complex_variant_name.txt"
  end
 
  def wget_hgvs4variation()
    hgvs_file = "hgvs4variation.txt"
    Dir.chdir("#{OUTPUT_DIR}") {
      system("wget ftp://ftp.ncbi.nlm.nih.gov/pub/clinvar/tab_delimited/#{hgvs_file}.gz")
      system("gunzip #{hgvs_file}.gz")
      system("awk -F'\t' 'BEGIN{OFS=\"\t\"}; {if($4 == \"-1\") print NR,$3,$4,$1,$5,$7,$11,$12;}' #{hgvs_file} | sort -k2,2n -k1,1n > #{@complex_variant_list_file}")
    }
  end

  def quote(str)
    return str.to_s.gsub(/(\\|\t|\n|\r|")/, '\\' => '\\\\', "\t" => '\\t', "\n" => '\\n', "\r" => '\\r', '"' => '\\"').inspect
  end

  def create_clinvar_label
    output_ttl_file = "#{OUTPUT_DIR}/clinvar_label.ttl"
    output_simple_clinvar_label(output_ttl_file)
    output_complex_clinvar_label(output_ttl_file)
  end
 
  def output_simple_clinvar_label (output_ttl_file_path)
    query_file = "#{@base_dir}/sparql/clinvar_label_from_simple_allele.rq"
    output_txt_file = "#{OUTPUT_DIR}/clinvar_label_from_simple_allele.txt"
    system(%Q[#{ISQL} #{ISQL_OPT} < #{query_file} > #{output_txt_file}])
    output_ttl_file = File.open(output_ttl_file_path, "w")
    output_ttl_file.puts "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> ."
    output_ttl_file.puts
    File.foreach(output_txt_file) do |line| 
      clinvar_id = line.split("^@").first.strip
      clinvar_label = line.split("^@").last.strip.chomp
      output_ttl_file.puts "<#{clinvar_id}>\trdfs:label\t#{quote(clinvar_label)} ."
    end
    output_ttl_file.flush
    output_ttl_file.close
  end

  def complex_name_hash
    complex_array = [] 
    File.foreach("#{OUTPUT_DIR}/#{@complex_variant_list_file}") do |line|
      columns = line.chomp.split("\t")
      hash = {
        hgvs_row_num: columns[0].strip,
        clinvar_id: columns[1].strip,
        gene_name: columns[3].strip,
        nuq_type: columns[4].strip,
        hgvs: columns[5].strip,
        used_for_naming: columns[6].strip,
        submitted: columns[7].strip
      }
      complex_array.push(hash)
    end
    complex_variant_hash = {}
    complex_array.group_by {|row|
      row[:clinvar_id]
    }.each do |clinvar, info|
      selected_item = nil
      if info.size == 1
        selected_item = info.first
      elsif
        info.each do |item|
          score = 0
          if item[:nuq_type].start_with?("coding")
            score = 100
          end
          if item[:used_for_naming] == "YES"
            score += 10
          end
          item[:score] = score
        end
        max_score = info.max_by {|item| item[:score] }[:score]
        high_score_items = info.find_all {|item| item[:score] == max_score }
        if high_score_items.size == 1
          selected_item = high_score_items.first
        else
          selected_item = high_score_items.max_by {|item| item[:hgvs_row_num].to_i }
        end
      end
      nuc = selected_item[:hgvs].split(":").first
      pos = selected_item[:hgvs].split(":")[1..-1].join()
      if selected_item[:gene_name] == "-"
        label = "#{nuc}:#{pos}"
      else
        label = "#{nuc}(#{selected_item[:gene_name]}):#{pos}"
      end
      complex_variant_hash[clinvar] = label 
    end
    complex_variant_hash
  end

  def output_complex_clinvar_label (output_ttl_file_path)
    query_file = "#{@base_dir}/sparql/clinvar_complex_variant_list.rq"
    output_txt_file = "#{OUTPUT_DIR}/clinvar_complex_variant_list.txt"
    system(%Q[#{ISQL} #{ISQL_OPT} < #{query_file} > #{output_txt_file}])
    output_ttl_file = File.open(output_ttl_file_path, "a")

    complex_name_hash_from_hgvs = complex_name_hash()
    unknown_label_ids = []
    File.foreach(output_txt_file) do |line| 
      clinvar_id = line.strip.chomp
      complex_label = complex_name_hash_from_hgvs[clinvar_id.split("/").last]
      unless complex_label.nil?
        output_ttl_file.puts "<#{clinvar_id}>\trdfs:label\t#{quote(complex_label)} ."
      else
        unknown_label_ids.push(clinvar_id.split("/").last)
      end
    end
    output_ttl_file.flush
    output_ttl_file.close
    #hgvsからlabel取得できなかったclinvar_idのlabelを組み立てて出力する
    get_label_from_alleles(unknown_label_ids, output_ttl_file_path)
  end

  def get_label_from_alleles(unknown_clinvar_list, output_ttl_file_path)
    return if unknown_clinvar_list.size == 0 
    clinvar_ids = unknown_clinvar_list.map {|id|
      "cvid:#{id}"
    }.join(" ")
    query_file = "#{@base_dir}/sparql/clinvar_unknwon_allele_label.rq"
    output_txt_file = "#{OUTPUT_DIR}/clinvar_unknwon_allele_label.txt"
    File.open("#{query_file}", "w") do |f|
      f.puts ERB.new(File.read("#{@base_dir}/sparql/clinvar_unknwon_allele_label.rq.erb")).result(binding)
    end
    system(%Q[#{ISQL} #{ISQL_OPT} < #{query_file} > #{output_txt_file}])

    allele_label_list = []
    File.foreach(output_txt_file) do |line| 
      clinvar_id = line.split("^@").first.strip
      allele_label = line.split("^@").last.strip.chomp
      hash = {clinvar_id: clinvar_id, allele_label: allele_label}
      #allele_label = "NM_001018077.1(NR3C1):c.66G>A (p.Glu22=)"
      #allele_label_nuc = "NM_001018077.1(NR3C1)"
      #split_pos_type = "c"
      #pos_label = "66G>A"
      #この形式をとる場合には同じclinvar_idの他のallele labelと次のようにグルーピング表示できる可能性があるため groupable = treuに設定
      #"NM_001018077.1(NR3C1):c.[66G>A; 68G>A]"
      groupable = false
      split_nucleotide = allele_label.split(":")
      if allele_label.size >= 2
        nuculeotide = split_nucleotide.first
        split_pos_type = split_nucleotide[1..-1].join("").split(".")
        if split_pos_type.size >= 2
          pos_type = split_pos_type.first
          hash[:group_key] = "#{nuculeotide}:#{pos_type}."
          hash[:pos_label] = split_pos_type[1..-1].join("").split("(").first.strip
          groupable = true
        end
      end
      hash[:groupable] = groupable
      allele_label_list.push(hash)
    end

    output_ttl_file = File.open(output_ttl_file_path, "a")
    allele_label_list.group_by {|row| row[:clinvar_id]}.each do |clinvar_id, allele_info|
      label = ""
      if allele_info.find {|item| item[:groupable] == false }.nil? #全てgrouping可能な形式であればgroupingする
        allele_info.group_by {|item| item[:group_key]}.each do |key, allele_labels|
          pos_label = allele_labels.map {|allele_label|
            allele_label[:pos_label]
          }.join("; ")
          label += "#{key}[#{pos_label}]"
        end 
      else
        label = allele_info.map {|item|
          item[allele_label]
        }.join(" ")
      end     
      output_ttl_file.puts "<#{clinvar_id}>\trdfs:label\t#{quote(label)} ."
    end

    output_ttl_file.flush
    output_ttl_file.close
  end

end

clinvar = ClinVarAdditionalData.new
clinvar.wget_hgvs4variation()
clinvar.create_clinvar_label()
